#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include "util.h"
#include "types.h"
#include "ArchTokens.h"
#include "ReadArch.h"
#include "OutputXML.h"

void Output_XML(char *FileName,
			  t_arch *Arch) {

	FILE* xmlFile;
	int direction;
	int i;

	xmlFile = my_fopen(FileName, "w");

	fprintf(xmlFile, "\t<architecture>\n");

	if (Arch->clb_grid.IsAuto == TRUE)
	{
		/* IF there is a layout aspect ratio then create it */
		fprintf(xmlFile, "\t\t<layout auto=\"%f\"/>\n", Arch->clb_grid.Aspect);
	}
	else
	{
		fprintf(xmlFile, "\t\t<layout auto=\"1.0\"/>\n");
	}

	fprintf(xmlFile, "\t\t<device>\n");

	fprintf(xmlFile, "\t\t\t<sizing R_minW_nmos=\"%f\" R_minW_pmos=\"%f\" ipin_mux_trans_size=\"%f\"/>\n", Arch->R_minW_nmos, Arch->R_minW_pmos, Arch->ipin_mux_trans_size);
	fprintf(xmlFile, "\t\t\t<timing C_ipin_cblock=\"%e\" T_ipin_cblock=\"%e\"/>\n", Arch->C_ipin_cblock, Arch->T_ipin_cblock);
	fprintf(xmlFile, "\t\t\t<area grid_logic_tile_area=\"%f\"/>\n", Arch->clb_area);

	fprintf(xmlFile, "\t\t\t<chan_width_distr>\n");
	fprintf(xmlFile, "\t\t\t\t<io width=\"%f\"/>\n", Arch->Chans.chan_width_io);
	if (Arch->Chans.chan_x_dist.type == UNIFORM)
		fprintf(xmlFile, "\t\t\t\t<x distr=\"uniform\" peak=\"%f\"/>\n", Arch->Chans.chan_x_dist.peak);
	else if (Arch->Chans.chan_x_dist.type == GAUSSIAN)
		fprintf(xmlFile, "\t\t\t\t<x distr=\"gaussian\" peak=\"%f\"/>\n", Arch->Chans.chan_x_dist.peak);
	else if (Arch->Chans.chan_x_dist.type == PULSE)
		fprintf(xmlFile, "\t\t\t\t<x distr=\"pulse\" peak=\"%f\"/>\n", Arch->Chans.chan_x_dist.peak);
	else if (Arch->Chans.chan_x_dist.type == DELTA)
		fprintf(xmlFile, "\t\t\t\t<x distr=\"delta\" peak=\"%f\"/>\n", Arch->Chans.chan_x_dist.peak);
	if (Arch->Chans.chan_y_dist.type == UNIFORM)
		fprintf(xmlFile, "\t\t\t\t<y distr=\"uniform\" peak=\"%f\"/>\n", Arch->Chans.chan_y_dist.peak);
	else if (Arch->Chans.chan_y_dist.type == GAUSSIAN)
		fprintf(xmlFile, "\t\t\t\t<y distr=\"gaussian\" peak=\"%f\"/>\n", Arch->Chans.chan_y_dist.peak);
	else if (Arch->Chans.chan_y_dist.type == PULSE)
		fprintf(xmlFile, "\t\t\t\t<y distr=\"pulse\" peak=\"%f\"/>\n", Arch->Chans.chan_y_dist.peak);
	else if (Arch->Chans.chan_y_dist.type == DELTA)
		fprintf(xmlFile, "\t\t\t\t<y distr=\"delta\" peak=\"%f\"/>\n", Arch->Chans.chan_y_dist.peak);
	fprintf(xmlFile, "\t\t\t</chan_width_distr>\n");

	if (Arch->Fs == 0)
		Arch->Fs = 3;
	if (Arch->SBType == WILTON)
		fprintf(xmlFile, "\t\t\t<switch_block type=\"wilton\" fs=\"%d\"/>\n", Arch->Fs);
	else if (Arch->SBType == SUBSET)
		fprintf(xmlFile, "\t\t\t<switch_block type=\"subset\" fs=\"%d\"/>\n", Arch->Fs);
	else if (Arch->SBType == UNIVERSAL)
		fprintf(xmlFile, "\t\t\t<switch_block type=\"universal\" fs=\"%d\"/>\n", Arch->Fs);
	else if (Arch->SBType == FULL)
		fprintf(xmlFile, "\t\t\t<switch_block type=\"full\" fs=\"%d\"/>\n", Arch->Fs);

	fprintf(xmlFile, "\t\t</device>\n");

	direction = Arch->Segments[0].Directionality;

	{
		struct s_switch *switch_traverse;

		fprintf(xmlFile, "\t\t<switchlist>\n");
		if (direction == UNI_DIRECTIONAL)
		{
			for (switch_traverse = Arch->Switches; switch_traverse != NULL; switch_traverse = switch_traverse->Next)
			{
				fprintf(xmlFile, "\t\t\t<switch type=\"mux\" name=\"%d\" R=\"%f\" Cin=\"%e\" Cout=\"%e\" Tdel=\"%e\" mux_trans_size=\"%f\"",
					switch_traverse->Index, switch_traverse->R, switch_traverse->Cin, switch_traverse->Cout, 
					switch_traverse->Tdel, switch_traverse->MuxTransSize);
				if (switch_traverse->BufSize != 0)
					fprintf(xmlFile, " buf_size=\"%f\"/>\n", switch_traverse->BufSize);
				else
					fprintf(xmlFile, "/>\n", switch_traverse->BufSize);
			}
		}
		else
		{
			for (switch_traverse = Arch->Switches; switch_traverse != NULL; switch_traverse = switch_traverse->Next)
			{
				if (switch_traverse->Buffered == TRUE)
				{
					fprintf(xmlFile, "\t\t\t<switch type=\"buffer\" name=\"%d\" R=\"%f\" Cin=\"%e\" Cout=\"%e\" Tdel=\"%e\" mux_trans_size=\"%f\"",
						switch_traverse->Index, switch_traverse->R, switch_traverse->Cin, switch_traverse->Cout, 
						switch_traverse->Tdel, switch_traverse->MuxTransSize) ;
					if (switch_traverse->BufSize != 0)
						fprintf(xmlFile, " buf_size=\"%f\"/>\n", switch_traverse->BufSize);
					else
						fprintf(xmlFile, "/>\n");
				}
				else
				{
					fprintf(xmlFile, "\t\t\t<switch type=\"pass_trans\" name=\"%d\" R=\"%f\" Cin=\"%e\" Cout=\"%e\" Tdel=\"%e\" mux_trans_size=\"%f\"",
						switch_traverse->Index, switch_traverse->R, switch_traverse->Cin, switch_traverse->Cout, 
						switch_traverse->Tdel, switch_traverse->MuxTransSize) ;
					if (switch_traverse->BufSize != 0)
						fprintf(xmlFile, " buf_size=\"%f\"/>\n", switch_traverse->BufSize);
					else
						fprintf(xmlFile, "/>\n");
				}
			}
		}
		fprintf(xmlFile, "\t\t</switchlist>\n");
	}

	{
		struct s_seg *segment_traverse;
		int count;

		fprintf(xmlFile, "\t\t<segmentlist>\n");
		if (direction == UNI_DIRECTIONAL)
		{
			for (segment_traverse = Arch->Segments; segment_traverse != NULL; segment_traverse = segment_traverse->Next)
			{
				fprintf(xmlFile, "\t\t\t<segment freq=\"%f\" length=\"%d\" type=\"unidir\" Rmetal=\"%f\" Cmetal=\"%e\">\n",
						segment_traverse->Freq, segment_traverse->Len, segment_traverse->Rmetal, segment_traverse->Cmetal);
				fprintf(xmlFile, "\t\t\t<mux name=\"%d\"/>\n", segment_traverse->WireSwitch);

				count = (int)(ceil(segment_traverse->Frac_sb*(segment_traverse->Len+1.0))) - 2;
				fprintf(xmlFile, "\t\t\t<sb type=\"pattern\">");
				for (i = 0; i < segment_traverse->Len+1; i++)
				{
					if (i == 0) 
						fprintf(xmlFile, "1");
					else if (i == segment_traverse->Len)
						fprintf(xmlFile, " 1");
					else if (count > 0)
					{
						fprintf(xmlFile, " 1");
						count --;
					}
					else
						fprintf(xmlFile, " 0");
				}
				fprintf(xmlFile, "</sb>\n");
				
				count = (int)(ceil(segment_traverse->Frac_sb*(segment_traverse->Len))) - 2;
				fprintf(xmlFile, "\t\t\t<cb type=\"pattern\">");
				for (i = 0; i < segment_traverse->Len; i++)
				{
					if (i == 0) 
						fprintf(xmlFile, "1");
					else if (i == segment_traverse->Len - 1)
						fprintf(xmlFile, " 1");
					else if (count > 0)
					{
						fprintf(xmlFile, " 1");
						count --;
					}
					else
						fprintf(xmlFile, " 0");
				}
				fprintf(xmlFile, "</cb>\n");

				fprintf(xmlFile, "\t\t\t</segment>\n");
			}
		}
		else
		{
			for (segment_traverse = Arch->Segments; segment_traverse != NULL; segment_traverse = segment_traverse->Next)
			{
				fprintf(xmlFile, "\t\t\t<segment length=\"%d\" type=\"bidir\" freq=\"%f\" Rmetal=\"%f\" Cmetal=\"%e\">\n",
						segment_traverse->Len, segment_traverse->Freq, segment_traverse->Rmetal, segment_traverse->Cmetal);
				fprintf(xmlFile, "\t\t\t<wire_switch name=\"%d\"/>\n", segment_traverse->WireSwitch);
				fprintf(xmlFile, "\t\t\t<opin_switch name=\"%d\"/>\n", segment_traverse->OPinSwitch);

				count = (int)(ceil(segment_traverse->Frac_sb*(segment_traverse->Len+1.0))) - 2;
				fprintf(xmlFile, "\t\t\t<sb type=\"pattern\">");
				for (i = 0; i < segment_traverse->Len+1; i++)
				{
					if (i == 0) 
						fprintf(xmlFile, "1");
					else if (i == segment_traverse->Len)
						fprintf(xmlFile, " 1");
					else if (count > 0)
					{
						fprintf(xmlFile, " 1");
						count --;
					}
					else
						fprintf(xmlFile, " 0");
				}
				fprintf(xmlFile, "</sb>\n");
				
				count = (int)(ceil(segment_traverse->Frac_sb*(segment_traverse->Len))) - 2;
				fprintf(xmlFile, "\t\t\t<cb type=\"pattern\">");
				for (i = 0; i < segment_traverse->Len; i++)
				{
					if (i == 0) 
						fprintf(xmlFile, "1");
					else if (i == segment_traverse->Len - 1)
						fprintf(xmlFile, " 1");
					else if (count > 0)
					{
						fprintf(xmlFile, " 1");
						count --;
					}
					else
						fprintf(xmlFile, " 0");
				}
				fprintf(xmlFile, "</cb>\n");

				fprintf(xmlFile, "\t\t\t</segment>\n");
			}
		}
		fprintf(xmlFile, "\t\t</segmentlist>\n");
	}

	fprintf(xmlFile, "\t\t<typelist>\n");

	fprintf(xmlFile, "\t\t\t<io capacity=\"%d\" t_inpad=\"%e\" t_outpad=\"%e\">\n",
			Arch->IO_Ratio, Arch->T_ipad, Arch->T_opad);
	if (Arch->Fc_Type == ABSOLUTE)
		fprintf(xmlFile, "\t\t\t\t<fc_in type=\"abs\">%f</fc_in>\n",
			Arch->Fc_Pad);
	else
		fprintf(xmlFile, "\t\t\t\t<fc_in type=\"frac\">%f</fc_in>\n",
			Arch->Fc_Pad);
	if (Arch->Fc_out_full_flex == TRUE)
		fprintf(xmlFile, "\t\t\t\t<fc_out type=\"full\"/>\n");
	else
		fprintf(xmlFile, "\t\t\t\t<fc_out type=\"frac\">%f</fc_out>\n", 
			Arch->Fc_Out);
	fprintf(xmlFile, "\t\t\t</io>\n");

	fprintf(xmlFile, "\t\t\t<type name=\".clb\">\n");
	fprintf(xmlFile, "\t\t\t\t<subblocks max_subblocks=\"%d\" max_subblock_inputs=\"%d\">\n",
			Arch->N, Arch->K);

	if (Arch->TSubblocks != NULL)
	{
		fprintf(xmlFile, "\t\t\t\t\t<timing>\n");
		fprintf(xmlFile, "\t\t\t\t\t\t<T_comb>\n");
	
		for (i = 0; i < Arch->K; i++)
		{
			fprintf(xmlFile, "\t\t\t\t\t\t\t<trow>%e</trow>\n",
					Arch->TSubblocks[0].T_comb);
		}
		fprintf(xmlFile, "\t\t\t\t\t\t</T_comb>\n");
	
		fprintf(xmlFile, "\t\t\t\t\t\t<T_seq_in>\n");
		fprintf(xmlFile, "\t\t\t\t\t\t\t<trow>%e</trow>\n",
				Arch->TSubblocks[0].T_seq_in - Arch->TSubblocks[0].T_comb);
		fprintf(xmlFile, "\t\t\t\t\t\t</T_seq_in>\n");
		
		fprintf(xmlFile, "\t\t\t\t\t\t<T_seq_out>\n");
		fprintf(xmlFile, "\t\t\t\t\t\t\t<trow>%e</trow>\n",
				Arch->TSubblocks[0].T_seq_out);
		fprintf(xmlFile, "\t\t\t\t\t\t</T_seq_out>\n");
		fprintf(xmlFile, "\t\t\t\t\t</timing>\n");
	}
	else
	{
		/*for (i = 0; i < Arch->K; i++)
		{
			fprintf(xmlFile, "\t\t\t\t\t\t\t<trow>0</trow>\n");
		}
		fprintf(xmlFile, "\t\t\t\t\t\t</T_comb>\n");
	
		fprintf(xmlFile, "\t\t\t\t\t\t<T_seq_in>\n");
		fprintf(xmlFile, "\t\t\t\t\t\t\t<trow>0</trow>\n");
		fprintf(xmlFile, "\t\t\t\t\t\t</T_seq_in>\n");
		
		fprintf(xmlFile, "\t\t\t\t\t\t<T_seq_out>\n");
		fprintf(xmlFile, "\t\t\t\t\t\t\t<trow>0</trow>\n");
		fprintf(xmlFile, "\t\t\t\t\t\t</T_seq_out>\n");
		fprintf(xmlFile, "\t\t\t\t\t</timing>\n");*/
	}
	fprintf(xmlFile, "\t\t\t\t</subblocks>\n");
	if (Arch->Fc_Type == ABSOLUTE)
		fprintf(xmlFile, "\t\t\t\t<fc_in type=\"abs\">%f</fc_in>\n",
			Arch->Fc_In);
	else
		fprintf(xmlFile, "\t\t\t\t<fc_in type=\"frac\">%f</fc_in>\n",
			Arch->Fc_In);
	if (Arch->Fc_out_full_flex == TRUE)
		fprintf(xmlFile, "\t\t\t\t<fc_out type=\"full\"/>\n");
	else
		fprintf(xmlFile, "\t\t\t\t<fc_out type=\"frac\">%f</fc_out>\n", 
			Arch->Fc_Out);

	{
		struct s_pin *pin_traverse;
		int index;
		int leftSideHasPins, rightSideHasPins, topSideHasPins, bottomSideHasPins;

		fprintf(xmlFile, "\t\t\t\t<pinclasses>\n");

		index = 0;
		fprintf(xmlFile, "\t\t\t\t\t<class type=\"in\">");
		for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
		{
			if (pin_traverse->IsGlobal == FALSE && pin_traverse->Type == RECEIVER)
				fprintf(xmlFile, "%d ", index);
			index++;
		}
		fprintf(xmlFile, "</class>\n");

		index = 0;
		fprintf(xmlFile, "\t\t\t\t\t<class type=\"out\">");
		for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
		{
			if (pin_traverse->IsGlobal == FALSE && pin_traverse->Type == DRIVER)
				fprintf(xmlFile, "%d ", index);
			index++;
		}
		fprintf(xmlFile, "</class>\n");

		index = 0;
		fprintf(xmlFile, "\t\t\t\t\t<class type=\"global\">");
		for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
		{
			if (pin_traverse->IsGlobal == TRUE)
				fprintf(xmlFile, "%d ", index);
			index++;
		}
		fprintf(xmlFile, "</class>\n");

		fprintf(xmlFile, "\t\t\t\t</pinclasses>\n");

		fprintf(xmlFile, "\t\t\t\t<pinlocations>\n");

		leftSideHasPins = rightSideHasPins = topSideHasPins = bottomSideHasPins = FALSE;		
		for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
		{
			if (pin_traverse->Side == LEFT)
				leftSideHasPins = TRUE;
			if (pin_traverse->Side == RIGHT)
				rightSideHasPins = TRUE;
			if (pin_traverse->Side == TOP)
				topSideHasPins = TRUE;
			if (pin_traverse->Side == BOTTOM)
				bottomSideHasPins = TRUE;
		}

		if(leftSideHasPins) {
			index = 0;
			fprintf(xmlFile, "\t\t\t\t\t<loc side=\"left\">");
			for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
			{
				if (pin_traverse->Side == LEFT)
					fprintf(xmlFile, "%d ", index);
				index++;
			}
			fprintf(xmlFile, "</loc>\n");
		}

		if(rightSideHasPins) {
			index = 0;
			fprintf(xmlFile, "\t\t\t\t\t<loc side=\"right\">");
			for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
			{
				if (pin_traverse->Side == RIGHT)
					fprintf(xmlFile, "%d ", index);
				index++;
			}
			fprintf(xmlFile, "</loc>\n");
		}

		if(topSideHasPins) {
			index = 0;
			fprintf(xmlFile, "\t\t\t\t\t<loc side=\"top\">");
			for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
			{
				if (pin_traverse->Side == TOP)
					fprintf(xmlFile, "%d ", index);
				index++;
			}
			fprintf(xmlFile, "</loc>\n");
		}

		if(bottomSideHasPins) {	
			index = 0;
			fprintf(xmlFile, "\t\t\t\t\t<loc side=\"bottom\">");
			for (pin_traverse = Arch->Pins; pin_traverse != NULL; pin_traverse = pin_traverse->Next)
			{
				if (pin_traverse->Side == BOTTOM)
					fprintf(xmlFile, "%d ", index);
				index++;
			}
			fprintf(xmlFile, "</loc>\n");
		}

		fprintf(xmlFile, "\t\t\t\t</pinlocations>\n");
	}

	fprintf(xmlFile, "\t\t\t\t<gridlocations>\n");
	fprintf(xmlFile, "\t\t\t\t\t<loc type=\"fill\" priority=\"1\"/>\n");
	fprintf(xmlFile, "\t\t\t\t</gridlocations>\n");

	fprintf(xmlFile, "\t\t\t\t<timing>\n");
	fprintf(xmlFile, "\t\t\t\t\t<tedge type=\"T_sblk_opin_to_sblk_ipin\">%e</tedge>\n",
			Arch->T_sblk_opin_to_sblk_ipin);
	fprintf(xmlFile, "\t\t\t\t\t<tedge type=\"T_fb_ipin_to_sblk_ipin\">%e</tedge>\n",
			Arch->T_clb_ipin_to_sblk_ipin);
	fprintf(xmlFile, "\t\t\t\t\t<tedge type=\"T_sblk_opin_to_fb_opin\">%e</tedge>\n",
			Arch->T_sblk_opin_to_clb_opin);
	fprintf(xmlFile, "\t\t\t\t</timing>\n");

	fprintf(xmlFile, "\t\t\t</type>\n");

	fprintf(xmlFile, "\t\t</typelist>\n");

	fprintf(xmlFile, "\t</architecture>\n");

	fclose(xmlFile);
}

